# -*- coding: utf-8 -*-

from . import bregman
from . import da
from .bregman import sinkhorn

# Author: Remi Flamary <remi.flamary@unice.fr>
#         Leo Gautheron <https://github.com/aje> 
#
# License: MIT License

__all__ = ["bregman", "da", "sinkhorn"]
